"""Miscellaneous functions (https://mariadb.com/kb/en/miscellaneous-functions/)."""

from sqlfactory.func.base import Function
from sqlfactory.statement import Statement


class GetLock(Function):
    """Tries to obtain a lock with a name."""

    def __init__(self, name: str | Statement, timeout: int | Statement) -> None:
        super().__init__("GET_LOCK", name, timeout)


class Inet6Aton(Function):
    """Converts an IPv6 address from its string representation to a binary string."""

    def __init__(self, ip: str | Statement) -> None:
        super().__init__("INET6_ATON", ip)


class Inet6Ntoa(Function):
    """Converts an IPv6 address from its binary string representation to a string."""

    def __init__(self, ip: str | Statement) -> None:
        super().__init__("INET6_NTOA", ip)


class InetAton(Function):
    """Converts an IPv4 address from its string representation to a number."""

    def __init__(self, ip: str | Statement) -> None:
        super().__init__("INET_ATON", ip)


class InetNtoa(Function):
    """Converts an IP number to a string representation."""

    def __init__(self, ip: int | Statement) -> None:
        super().__init__("INET_NTOA", ip)


class IsFreeLock(Function):
    """Checks whether a named lock is free."""

    def __init__(self, name: str | Statement) -> None:
        super().__init__("IS_FREE_LOCK", name)


class IsIpv4(Function):
    """Checks whether a string is an IPv4 address."""

    def __init__(self, ip: str | Statement) -> None:
        super().__init__("IS_IPV4", ip)


class IsIpv4Compat(Function):
    """Checks whether IPv6 address is a valid IPv4-compatible address."""

    def __init__(self, ip: str | Statement) -> None:
        super().__init__("IS_IPV4_COMPAT", ip)


class IsIpv4Mapped(Function):
    """Checks whether IPv6 is an IPv4-mapped address."""

    def __init__(self, ip: str) -> None:
        super().__init__("IS_IPV4_MAPPED", ip)


class IsIpv6(Function):
    """Checks whether a string is an IPv6 address."""

    def __init__(self, ip: str | Statement) -> None:
        super().__init__("IS_IPV6", ip)


class IsUsedLock(Function):
    """Checks whether a named lock is in use."""

    def __init__(self, name: str | Statement) -> None:
        super().__init__("IS_USED_LOCK", name)


class MasterGtidWait(Function):
    """Waits until the slave reaches a specified GTID position."""

    def __init__(self, gtid_set: str | Statement, timeout: int | Statement | None = None) -> None:
        if timeout is not None:
            super().__init__("MASTER_GTID_WAIT", gtid_set, timeout)
        else:
            super().__init__("MASTER_GTID_WAIT", gtid_set)


class MasterPosWait(Function):
    """Waits until the slave reaches a specified binary log position."""

    def __init__(self, log_file: str | Statement, log_pos: int | Statement, timeout: int | Statement | None = None) -> None:
        if timeout is not None:
            super().__init__("MASTER_POS_WAIT", log_file, log_pos, timeout)
        else:
            super().__init__("MASTER_POS_WAIT", log_file, log_pos)


class ReleaseAllLocks(Function):
    """Releases all named locks."""

    def __init__(self) -> None:
        super().__init__("RELEASE_ALL_LOCKS")


class ReleaseLock(Function):
    """Releases a named lock."""

    def __init__(self, name: str | Statement) -> None:
        super().__init__("RELEASE_LOCK", name)


class Sleep(Function):
    """Sleeps for a specified number of seconds."""

    def __init__(self, seconds: int | Statement) -> None:
        super().__init__("SLEEP", seconds)


class SysGuid(Function):
    """Returns a globally unique identifier."""

    def __init__(self) -> None:
        super().__init__("SYS_GUID")


class Uuid(Function):
    """Returns a universally unique identifier."""

    def __init__(self) -> None:
        super().__init__("UUID")


class UuidShort(Function):
    """Returns a short universally unique identifier."""

    def __init__(self) -> None:
        super().__init__("UUID_SHORT")



# noqa: A005  # shadows built-in select
"""SELECT statement builder."""

from sqlfactory.mixins.join import CrossJoin, InnerJoin, Join, LeftJoin, LeftOuterJoin, RightJoin, RightOuterJoin
from sqlfactory.select.aliased import Aliased, SelectColumn
from sqlfactory.select.column_list import ColumnList
from sqlfactory.select.cte import With
from sqlfactory.select.select import SELECT, Select
from sqlfactory.select.union import (
    EXCEPT,
    EXCEPT_ALL,
    EXCEPT_DISTINCT,
    INTERSECT,
    INTERSECT_ALL,
    INTERSECT_DISTINCT,
    UNION,
    UNION_ALL,
    UNION_DISTINCT,
    Except,
    ExceptAll,
    ExceptDistinct,
    Intersect,
    IntersectAll,
    IntersectDistinct,
    Union,
    UnionAll,
    UnionDistinct,
)

__all__ = [
    "EXCEPT",
    "EXCEPT_ALL",
    "EXCEPT_DISTINCT",
    "INTERSECT",
    "INTERSECT_ALL",
    "INTERSECT_DISTINCT",
    "SELECT",
    "UNION",
    "UNION_ALL",
    "UNION_DISTINCT",
    "Aliased",
    "ColumnList",
    "CrossJoin",
    "Except",
    "ExceptAll",
    "ExceptDistinct",
    "InnerJoin",
    "Intersect",
    "IntersectAll",
    "IntersectDistinct",
    "Join",
    "LeftJoin",
    "LeftOuterJoin",
    "RightJoin",
    "RightOuterJoin",
    "Select",
    "SelectColumn",
    "Union",
    "UnionAll",
    "UnionDistinct",
    "With",
]



"""Simple binary comparison conditions."""

from typing import Any, TypeAlias

from sqlfactory.condition.base import ConditionBase, StatementOrColumn
from sqlfactory.statement import Statement


class SimpleCondition(ConditionBase):
    """
    Simple condition comparing one column with given value, using specified operator.

    **Note:** `SimpleCondition` and all other condition functions are expecting first argument to be column and second argument
    to be value. If you want to have column on the right side or value on the left, you must explicitely use `Column()` function.
    And if you want literal value on the left side, you must explicitely use `Value()` function.
    """

    def __init__(self, column: StatementOrColumn, operator: str, value: Statement | Any) -> None:
        # pylint: disable=duplicate-code   # It does not make sense to generalize two-row statement used on two places.
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param operator: Operator to use for comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__()

        if not isinstance(column, Statement):
            # pylint: disable=import-outside-toplevel,cyclic-import
            from sqlfactory.entities import Column

            column = Column(column)

        self._column = column
        self._operator = operator
        self._value = value

    @property
    def args(self) -> list[Any]:
        args = [*self._column.args]

        if isinstance(self._value, Statement):
            args.extend(self._value.args)

        elif not isinstance(self._value, Statement):
            args.append(self._value)

        return args

    def __str__(self) -> str:
        if isinstance(self._value, Statement):
            return f"{self._column!s} {self._operator} {self._value!s}"

        return f"{self._column!s} {self._operator} {self.dialect.placeholder}"

    def __bool__(self) -> bool:
        return True


class Equals(SimpleCondition):
    """
    Equality condition (`==`). You can also use shorthand alias `Eq`.

    **Note:** First argument to `Equals` (or `Eq`) is expected to be column, while second argument is value. So to compare
    two columns, you must use `Column` instances as second argument (`Eq("column1", Column("column2"))` to produce
    ``` `column1` = `column2` ```). Likewise, to provide literal value as the first argument, you must use `Value` instance
    (`Eq(Value(10), "column")` to produce ``` %s = `column` ``` (with args `[10]`)).

    ```Eq("column1", "column2")``` produces ``` `column1` = %s ``` with arguments ```["column2"]```.

    You can also use column's operator overloading:

    ```python
    c = Column("column")
    c == "value"  # == Eq("column", "value")
    c == None     # == Eq("column", None)
    c == Now()    # == Eq("column", Now())
    ```

    Examples:

    - Simple comparison of column to value
        ```python
        # `column` = <value>
        Eq("column", "value")
        "`column` = %s", ["value"]
        ```
    - Comparison column to None
        ```python
        # `column` IS NULL
        Eq("column", None)
        "`column` IS NULL"
        ```
    - Comparison of generic statement to value
        ```python
        # <statement> = <value>
        Eq(Date(), "2021-01-01")
        "DATE() = %s", ["2021-01-01"]
        ```
    - Comparison of generic statement to None
        ```python
        # <statement> IS NULL
        Eq(Date(), None)
        "DATE() IS NULL"
        ```
    - Comparison of statement to statement
        ```python
        # <statement> = <statement>
        Eq(Date(), Now())
        # "DATE() = NOW()"
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | None | Statement) -> None:
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        if value is None:
            super().__init__(column, "IS", value)
        else:
            super().__init__(column, "=", value)


class NotEquals(SimpleCondition):
    """
    Not equality condition (`!=`). You can also use shorthand alis `Ne`.

    **Note:** First argument to `NotEquals` (or `Ne`) is expected to be column, while second argument is value. So to compare
    two columns, you must use `Column` instances as second argument (`Ne("column1", Column("column2"))` to produce
    ``` `column1` != `column2` ```). Likewise, to provide literal value as the first argument, you must use `Value` instance
    (`Ne(Value(10), "column")` to produce ``` %s != `column` ``` (with args `[10]`)).

    ```Ne("column1", "column2")``` produces ``` `column1` != %s ``` with arguments ```["column2"]```.

    You can also use column's operator overloading:

    ```python
    c = Column("column")
    c != "value"  # == Ne("column", "value")
    c != None     # == Ne("column", None)
    c != Now()    # == Ne("column", Now())
    ```

    Examples:

    - Column not equals value
        ```python
        # `column` != <value>
        Ne("column", "value")
        "`column` != %s", ["value"]
        ```
    - Statement not equals value
        ```python
        # <statement> != <value>
        Ne(Date(), "2021-01-01")
        "DATE() != %s", ["2021-01-01"]
        ```
    - Column is not None
        ```python
        # `column` IS NOT NULL
        Ne("column", None)
        "`column` IS NOT NULL"
        ```
    - Statement is not None
        ```python
        # <statement> IS NOT NULL
        Ne(Date(), None)
        "DATE() IS NOT NULL"
        ```
    - Statement not equals statement
        ```python
        # <statement> != <statement>
        Ne(Date(), Now())
        "DATE() != NOW()"
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | None | Statement) -> None:
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        if value is None:
            super().__init__(column, "IS NOT", value)
        else:
            super().__init__(column, "!=", value)


class GreaterThanOrEquals(SimpleCondition):
    """
    Greater than or equal condition (`>=`). You can also use shorthand alis `Ge`.

    **Note:** First argument to `GreaterThanOrEquals` (or `Ge`) is expected to be column, while second argument is value.
    So to compare two columns, you must use `Column` instances as second argument (`Ge("column1", Column("column2"))` to produce
    ``` `column1` >= `column2` ```). Likewise, to provide literal value as the first argument, you must use `Value` instance
    (`Ge(Value(10), "column")` to produce ``` %s >= `column` ``` (with args `[10]`)).

    ```Ge("column1", "column2")``` produces ``` `column1` >= %s ``` with arguments ```["column2"]```.

    You can also use column's operator overloading:

    ```python
    c = Column("column")
    c >= "value"  # == Ge("column", "value")
    c >= None     # == Ge("column", None)
    c >= Now()    # == Ge("column", Now())
    ```

    Examples:

    - Column is greater or equals to value:
        ```python
        # `column` >= <value>
        Ge("column", 10)
        "`column` >= %s", [10]
        ```
    - Statement is greater or equals to value:
        ```python
        # <statement> >= <value>
        Ge(Date(), "2021-01-01")
        "DATE() >= %s", ["2021-01-01"]
        ```
    - Column is greater or equals to other column
        ```python
        # `column1` >= `column2`
        Ge("column1", Column("column2"))
        "`column1` >= `column2`"
        ```
    - Column is greater or equals to statement
        ```python
        # `column` >= <statement>
        Ge("column", Now())
        "`column` >= NOW()"
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement) -> None:
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__(column, ">=", value)


class GreaterThan(SimpleCondition):
    """
    Greater than condition (`>`). You can also use shorthand alis `Gt`.

    **Note:** First argument to `GreaterThan` (or `Ge`) is expected to be column, while second argument is value.
    So to compare two columns, you must use `Column` instances as second argument (`Gt("column1", Column("column2"))` to produce
    ``` `column1` > `column2` ```). Likewise, to provide literal value as the first argument, you must use `Value` instance
    (`Gt(Value(10), "column")` to produce ``` %s > `column` ``` (with args `[10]`)).

    ```Gt("column1", "column2")``` produces ``` `column1` > %s ``` with arguments ```["column2"]```.

    You can also use column's operator overloading:

    ```python
    c = Column("column")
    c > "value"  # == Gt("column", "value")
    c > None     # == Gt("column", None)
    c > Now()    # == Gt("column", Now())
    ```

    Examples:

    - Column is greater than value:
        ```python
        # `column` > <value>
        Gt("column", 10)
        "`column` > %s, [10]
        ```
    - Statement is greater than value:
        ```python
        # <statement> > <value>
        Gt(Date(), "2021-01-01")
        "DATE() > %s", ["2021-01-01"]
        ```
    - Column is greater than other column
        ```python
        # `column1` > `column2`
        Gt("column1", Column("column2"))
        "`column1` > `column2`"
        ```
    - Column is greater than statement
        ```python
        # `column` > <statement>
        Gt("column", Now())
        "`column` > NOW()"
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement) -> None:
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__(column, ">", value)


class LessThanOrEquals(SimpleCondition):
    """
    Less than or equal condition (`<=`). You can also use shorthand alis `Le`.

    **Note:** First argument to `LessThanOrEquals` (or `Le`) is expected to be column, while second argument is value.
    So to compare two columns, you must use `Column` instances as second argument (`Le("column1", Column("column2"))` to produce
    ``` `column1` <= `column2` ```). Likewise, to provide literal value as the first argument, you must use `Value` instance
    (`Le(Value(10), "column")` to produce ``` %s <= `column` ``` (with args `[10]`)).

    ```Le("column1", "column2")``` produces ``` `column1` <= %s ``` with arguments ```["column2"]```.

    You can also use column's operator overloading:

    ```python
    c = Column("column")
    c <= "value"  # == Le("column", "value")
    c <= None     # == Le("column", None)
    c <= Now()    # == Le("column", Now())
    ```

    Examples:

    - Column is lower or equal to value:
        ```python
        # `column` <= <value>
        Le("column", 10)
        "`column` <= %s", [10]
        ```
    - Statement is lower or equal to value
        ```python
        <statement> <= <value>
        Le(Date(), "2021-01-01")
        "DATE() <= %s", ["2021-01-01"]
        ```
    - Column is lower or equal to other column
        ```python
        # `column1` <= `column2`
        Le("column1", Column("column2"))
        "`column1` <= `column2`"
        ```
    - Column is lower or equal to statement
        ```python
        # `column` <= <statement>
        Le("column", Now())
        "`column` <= NOW()"
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement) -> None:
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__(column, "<=", value)


class LessThan(SimpleCondition):
    """
    Less than condition (`<`). You can also use shorthand alis `Lt`.

    **Note:** First argument to `LessThan` (or `Lt`) is expected to be column, while second argument is value.
    So to compare two columns, you must use `Column` instances as second argument (`Lt("column1", Column("column2"))` to produce
    ``` `column1` < `column2` ```). Likewise, to provide literal value as the first argument, you must use `Value` instance
    (`Lt(Value(10), "column")` to produce ``` %s < `column` ``` (with args `[10]`)).

    ```Lt("column1", "column2")``` produces ``` `column1` < %s ``` with arguments ```["column2"]```.

    You can also use column's operator overloading:

    ```python
    c = Column("column")
    c < "value"  # == Lt("column", "value")
    c < None     # == Lt("column", None)
    c < Now()    # == Lt("column", Now())
    ```

    Examples:

    - Column is lower than value:
        ```python
        # `column` < <value>
        Lt("column", 10)
        "`column` < %s", [10]
        ```
    - Statement is lower than value:
        ```python
        # <statement> < <value>
        Lt(Date(), "2021-01-01")
        "DATE() < %s", ["2021-01-01"]
        ```
    - Column is lower than other column
        ```python
        # `column1` < `column2`
        Lt("column1", Column("column2"))
        "`column1` < `column2`"
        ```
    - Column is lower than statement
        ```python
        # `column` < <statement>
        Lt("column", Now())
        "`column` < NOW()"
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement) -> None:
        """
        :param column: Column to compare (string or instance of Column) or Statement to use on left side of comparison.
        :param value: Value to compare column value to (or statement to use on right side of comparison).
        """
        super().__init__(column, "<", value)


# Convenient aliases for shorter code.
Eq: TypeAlias = Equals
Ge: TypeAlias = GreaterThanOrEquals
Gt: TypeAlias = GreaterThan
Le: TypeAlias = LessThanOrEquals
Lt: TypeAlias = LessThan
Ne: TypeAlias = NotEquals



"""WHERE mixin for query generator."""

from __future__ import annotations

from typing import Any, Self

from sqlfactory.condition.base import And, ConditionBase


class WithWhere:
    """
    Mixin to provide `WHERE` support for query generator.

    Example:

    >>> from sqlfactory import Select, And, Equals
    >>> Select(where=And(Equals("id", 1), Equals("status", "active")))

    >>> from sqlfactory import Update, And, Equals
    >>> Update(where=And(Equals("id", 1), Equals("status", "active")))

    >>> from sqlfactory import Delete, And, Equals
    >>> Delete(where=And(Equals("id", 1), Equals("status", "active")))

    """

    def __init__(self, *args: Any, where: ConditionBase | None = None, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._where = where

    def where(self, condition: ConditionBase) -> Self:
        """
        Set `WHERE` condition for query. If you chain multiple where calls together, they are joined by `And` condition.
        You can manually specify joining condition by use one of `sqlfactory.And` or `sqlfactory.Or` compound classes to chain
        multiple conditions.

        Example:

        >>> from sqlfactory import Select, And, Equals
        >>> sel = Select().where(And(Equals("id", 1), Equals("status", "active")))

        **Note:** `Eq()` and all other condition functions are expecting first argument to be column and second argument to be
        value. If you want to have column on the right side or value on the left, you must explicitely use `Column()` function.
        And if you want literal value on the left side, you must explicitely use `Value()` function.

        :param condition: Condition to be used in WHERE clause.
        """
        if self._where is not None:
            if isinstance(self._where, And):
                self._where.append(condition)
            else:
                self._where = And(self._where, condition)
        else:
            self._where = condition

        return self

    def WHERE(self, condition: ConditionBase) -> Self:  # pylint: disable=invalid-name
        """Alias for `WithWhere.where()` to be more SQL-like with all capitals."""
        return self.where(condition)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Collection, Self, overload

from sqlfactory.condition import ConditionBase
from sqlfactory.entities import Table
from sqlfactory.statement import Statement

if TYPE_CHECKING:
    from sqlfactory.select import Select  # pragma: no cover


class Join(Statement):
    """
    Produces `JOIN` statement

    Usage:

    >>> from sqlfactory import Join, Eq
    >>> Join("table2", on=Eq("table1.id", "table2.id"))
    >>> "JOIN `table2` ON `table1`.`id` = `table2`.`id`"

    >>> from sqlfactory import Join, Eq
    >>> Join("table2", on=Eq("table1.id", "t2.id"), alias="t2")
    >>> "JOIN `table2` AS `t2` ON `table1`.`id` = `t2`.`id`"

    With subquery:

    >>> from sqlfactory import Join, Eq, Select, Column
    >>> from sqlfactory.func.agg import Sum
    >>> Join(
    ...     Select("table2.product_id", Sum("price"), table="table2", where=Eq("table2.enabled", True)),
    ...     alias="t2",
    ...     on=Eq("t2.product_id", Column("table.id"))
    ... )
    >>> "JOIN ("
    ... "SELECT `table2`.`product_id`, SUM(`price`) "
    ... "FROM `table2` "
    ... "WHERE `table2`.`enabled` = %s"
    ... ") AS `t2` ON `t2`.`product_id` = `table`.`id`"

    Note that for subquery join, alias must be always specified.
    """

    def __init__(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> None:
        """
        :param table: Table to be joined
        :param on: ON condition
            **Note:** `Eq()` and all other condition functions are expecting first argument to be column and second argument to
            be value. If you want to have column on the right side or value on the left, you must explicitly use `Column()`
            function. And if you want literal value on the left side, you must explicitly use `Value()` function.
        :param alias: Optional alias of the joined table.
        """
        super().__init__()

        if isinstance(table, str):
            table = Table(table)

        self.table = table
        """Table to be joined"""

        self.on = on
        """ON join condition"""

        self.alias = alias
        """Optional alias for the joined table."""

        from sqlfactory.select import Select  # pylint: disable=import-outside-toplevel, cyclic-import

        if isinstance(self.table, Select) and not self.alias:
            raise AttributeError("When joining a subselect, alias must be specified.")

    @property
    def join_spec(self) -> str:
        """
        Returns the JOIN type itself for generation of SQL query.
        @private
        """
        return "JOIN"

    def __str__(self) -> str:
        from sqlfactory.select import Select  # pylint: disable=import-outside-toplevel, cyclic-import

        if self.alias:
            if isinstance(self.table, Select):
                table = f"({self.table!s}) AS `{self.alias}`"
            else:
                table = f"{self.table!s} AS `{self.alias}`"
        else:
            table = str(self.table)

        if self.on:
            return f"{self.join_spec} {table} ON {self.on!s}"

        return f"{self.join_spec} {table}"

    @property
    def args(self) -> list[Any]:
        """Argument values of the JOIN statement."""
        return [*(self.table.args if isinstance(self.table, Statement) else []), *(self.on.args if self.on else [])]


class LeftJoin(Join):
    """
    Produces `LEFT JOIN` statement

    Usage:

    >>> from sqlfactory import LeftJoin, Eq, Column
    >>> LeftJoin("table2", on=Eq("table1.id", Column("table2.id")))
    >>> "LEFT JOIN `table2` ON `table1`.`id` = `table2`.`id`"
    """

    @property
    def join_spec(self) -> str:
        return "LEFT JOIN"


class LeftOuterJoin(Join):
    """
    Produces `LEFT OUTER JOIN` statement

    Usage:

    >>> from sqlfactory import LeftOuterJoin, Eq, Column
    >>> LeftOuterJoin("table2", on=Eq("table1.id", Column("table2.id")))
    >>> "LEFT OUTER JOIN `table2` ON `table1`.`id` = `table2`.`id`"
    """

    @property
    def join_spec(self) -> str:
        return "LEFT OUTER JOIN"


class RightJoin(Join):
    """
    Produces `RIGHT JOIN` statement

    Usage:

    >>> from sqlfactory import RightJoin, Eq, Column
    >>> RightJoin("table2", on=Eq("table1.id", Column("table2.id")))
    >>> "RIGHT JOIN `table2` ON `table1`.`id` = `table2`.`id`"
    """

    @property
    def join_spec(self) -> str:
        return "RIGHT JOIN"


class RightOuterJoin(Join):
    """
    Produces `RIGHT OUTER JOIN` statement

    Usage:

    >>> from sqlfactory import RightOuterJoin, Eq, Column
    >>> RightOuterJoin("table2", on=Eq("table1.id", Column("table2.id")))
    >>> "RIGHT OUTER JOIN `table2` ON `table1`.`id` = `table2`.`id`"
    """

    @property
    def join_spec(self) -> str:
        return "RIGHT OUTER JOIN"


class InnerJoin(Join):
    """
    Produces `INNER JOIN` statement

    Usage:

    >>> from sqlfactory import InnerJoin, Eq, Column
    >>> InnerJoin("table2", on=Eq("table1.id", Column("table2.id")))
    >>> "INNER JOIN `table2` ON `table1`.`id` = `table2`.`id`"
    """

    @property
    def join_spec(self) -> str:
        return "INNER JOIN"


class CrossJoin(Join):
    """
    Produces `CROSS JOIN` statement

    Usage:

    >>> from sqlfactory import CrossJoin, Eq, Column
    >>> CrossJoin("table2", on=Eq("table1.id", Column("table2.id")))
    >>> "CROSS JOIN `table2` ON `table1`.`id` = `table2`.`id`"
    """

    @property
    def join_spec(self) -> str:
        return "CROSS JOIN"


class WithJoin:
    """Mixin to provide JOIN support for query generator."""

    def __init__(self, *args: Any, join: Collection[Join] | None = None, **kwargs: Any) -> None:
        """
        :param join: List of JOIN clauses
        """
        super().__init__(*args, **kwargs)
        self._join = list(join) if join is not None else []

    def _append_join(self, join: Join) -> Self:
        """Append join to list of joins."""
        if not self._join:
            self._join = []

        if join not in self._join:
            self._join.append(join)

        return self

    @overload
    def join(self, join: Join, /) -> Self:
        """Append JOIN clause to the query (any Join instance)."""

    @overload
    def join(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        """
        Append JOIN clause to the query.
        JOIN `table` AS <alias> ON (<condition>)

        :param table: Table to be joined
        :param on: ON condition
            **Note:** `Eq()` and all other condition functions are expecting first argument to be column and second argument to
            be value. If you want to have column on the right side or value on the left, you must explicitly use `Column()`
            function. And if you want literal value on the left side, you must explicitly use `Value()` function.
        :param alias: Optional alias of the joined table.
        """

    def join(self, table: str | Table | Join | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        """
        Append JOIN clause to the query.
        JOIN `table` AS <alias> ON (<condition>)

        :param table: Table to be joined (or instance of the Join class)
        :param on: ON condition (when first argument is not instance of Join class)
            **Note:** `Eq()` and all other condition functions are expecting first argument to be column and second argument to
            be value. If you want to have column on the right side or value on the left, you must explicitly use `Column()`
            function. And if you want literal value on the left side, you must explicitly use `Value()` function.
        :param alias: Optional alias of the joined table, if table is not instance of the Join class.
        """
        if isinstance(table, Join):
            if on is not None or alias is not None:
                raise AttributeError("When passing Join instance directly, on or alias attributes cannot be specified.")

            return self._append_join(table)

        return self._append_join(Join(table, on, alias))

    @overload
    def JOIN(self, join: Join, /) -> Self:  # pylint: disable=invalid-name
        """Alias for join() to be more SQL-like with all capitals."""

    @overload
    def JOIN(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        # pylint: disable=invalid-name
        """Alias for join() to be more SQL-like with all capitals."""

    def JOIN(self, table: str | Table | Select | Join, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        # pylint: disable=invalid-name
        """Alias for join() to be more SQL-like with all capitals."""
        return self.join(table, on, alias)  # type: ignore[arg-type]  # mypy searches in overloads

    def left_join(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        """
        Append LEFT JOIN clause to the query.

        :param table: Table to be joined
        :param on: ON condition
            **Note:** `Eq()` and all other condition functions are expecting first argument to be column and second argument to
            be value. If you want to have column on the right side or value on the left, you must explicitly use `Column()`
            function. And if you want literal value on the left side, you must explicitly use `Value()` function.
        :param alias: Optional alias of the joined table.
        """
        return self.join(LeftJoin(table, on, alias))

    def LEFT_JOIN(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        # pylint: disable=invalid-name
        """Alias for left_join() to be more SQL-like with all capitals."""
        return self.left_join(table, on, alias)

    def right_join(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        """
        Append RIGHT JOIN clause to the query.

        :param table: Table to be joined
        :param on: ON condition
            **Note:** `Eq()` and all other condition functions are expecting first argument to be column and second argument to
            be value. If you want to have column on the right side or value on the left, you must explicitly use `Column()`
            function. And if you want literal value on the left side, you must explicitly use `Value()` function.
        :param alias: Optional alias of the joined table.
        """
        return self.join(RightJoin(table, on, alias))

    def RIGHT_JOIN(self, table: str | Table | Select, on: ConditionBase | None = None, alias: str | None = None) -> Self:
        # pylint: disable=invalid-name
        """Alias for right_join() to be more SQL-like with all capitals."""
        return self.right_join(table, on, alias)



"""BETWEEN condition generator"""

from typing import Any, NoReturn

from sqlfactory.condition.base import ConditionBase, StatementOrColumn
from sqlfactory.entities import Column
from sqlfactory.statement import Statement


class Between(ConditionBase):
    # pylint: disable=duplicate-code  # It does not make sense to generalize two-row statement used on two places.
    """
    Provides generation for following syntax:

    - ``` `column` BETWEEN <lower_bound> AND <upper_bound>```
    - ``` `column` NOT BETWEEN <lower_bound> AND <upper_bound>```
    - ```<statement> BETWEEN <lower_bound> AND <upper_bound>```
    - ```<statement> NOT BETWEEN <lower_bound> AND <upper_bound>```

    Usage:

    >>> Between('column', 1, 10)
    >>> "`column` BETWEEN 1 AND 10"

    >>> Between('column', 1, 10, negative=True)
    >>> "`column` NOT BETWEEN 1 AND 10"

    >>> Between(Column('c1') + Column('c2'), 1, 10)
    >>> "(`c1` + `c2`) BETWEEN 1 AND 10"

    >>> Between(Column('c1') + Column('c2'), Column('c3') + Column('c4'), Column('c5') + Column('c6'))
    >>> "(`c1` + `c2`) BETWEEN (`c3` + `c4`) AND (`c5` + `c6`)"

    """

    def __init__(
        self, column: StatementOrColumn, lower_bound: Any | Statement, upper_bound: Any | Statement, *, negative: bool = False
    ) -> None:
        """
        :param column: Column to be compared.
        :param lower_bound: Lower inclusive bound of matching value
        :param upper_bound: Upper inclusive bound of matching value
        :param negative: Whether to negate the condition.
        """
        super().__init__()

        if not isinstance(column, Statement):
            column = Column(column)

        self._column = column
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound
        self._negative = negative

    def __str__(self) -> str:
        lower_bound_s = self.dialect.placeholder
        upper_bound_s = self.dialect.placeholder

        if isinstance(self._lower_bound, Statement):
            lower_bound_s = str(self._lower_bound)

        if isinstance(self._upper_bound, Statement):
            upper_bound_s = str(self._upper_bound)

        return f"{self._column!s} {'NOT ' if self._negative else ''}BETWEEN {lower_bound_s} AND {upper_bound_s}"

    @property
    def args(self) -> list[Any]:
        """
        Return argument values of the condition.
        """

        args = []

        if isinstance(self._column, Statement):
            args.extend(self._column.args)

        if isinstance(self._lower_bound, Statement):
            args.extend(self._lower_bound.args)
        else:
            args.append(self._lower_bound)

        if isinstance(self._upper_bound, Statement):
            args.extend(self._upper_bound.args)
        else:
            args.append(self._upper_bound)

        return args

    def __bool__(self) -> bool:
        return True

    def __invert__(self) -> "Between":
        """
        Allows using the `~` operator to negate the BETWEEN condition, converting it to a NOT BETWEEN condition.
        Note: Cannot use ~ operator on NotBetween conditions.
        """
        return NotBetween(self._column, self._lower_bound, self._upper_bound)


class NotBetween(Between):
    """
    Provides generation for following syntax:

    - ``` `column` NOT BETWEEN <lower_bound> AND <upper_bound>```
    - ```<statement> NOT BETWEEN <lower_bound> AND <upper_bound>```

    This is a dedicated class for NOT BETWEEN conditions, which is equivalent to using Between with negative=True.
    It provides a more intuitive API for NOT BETWEEN conditions.

    Usage:

    >>> NotBetween('column', 1, 10)
    >>> "`column` NOT BETWEEN 1 AND 10"

    >>> NotBetween(Column('c1') + Column('c2'), 1, 10)
    >>> "(`c1` + `c2`) NOT BETWEEN 1 AND 10"

    >>> NotBetween(Column('c1') + Column('c2'), Column('c3') + Column('c4'), Column('c5') + Column('c6'))
    >>> "(`c1` + `c2`) NOT BETWEEN (`c3` + `c4`) AND (`c5` + `c6`)"
    """

    def __init__(self, column: StatementOrColumn, lower_bound: Any | Statement, upper_bound: Any | Statement) -> None:
        """
        :param column: Column to be compared.
        :param lower_bound: Lower inclusive bound of matching value
        :param upper_bound: Upper inclusive bound of matching value
        """
        super().__init__(column, lower_bound, upper_bound, negative=True)

    def __invert__(self) -> NoReturn:
        """
        Allows using the `~` operator to negate the BETWEEN condition, converting it to a NOT BETWEEN condition.
        Note: Cannot use ~ operator on NotBetween conditions.
        """
        raise TypeError("Cannot use ~ operator on NotBetween conditions")



from typing import Any, Collection, Self

from sqlfactory.dialect import SQLDialect
from sqlfactory.entities import Column, Table
from sqlfactory.execute import ExecutableStatement
from sqlfactory.select.select import Select
from sqlfactory.select.union import Union


class With(ExecutableStatement):
    """
    Common Table Expression (CTE) statement.

    Usage:

    >>> from sqlfactory import With, Select, Gt, Eq, Column
    >>> With('t').as_(
    ...     Select('a', table='t1', where=Gt('b', 1))
    ... ).select(
    ...     Select('a', 'b', 'c', table=['t2', 't'], where=Eq('t2.c', Column('t.a'))
    ... )
    ... "WITH `t` AS (SELECT `a` FROM `t1` WHERE `b` > 1) SELECT `a`, `b`, `c` FROM `t2`, `t` WHERE `t2`.`c` = `t`.`a`"

    Recursive CTE:

    >>> from sqlfactory import With, Select, Aliased, Union, Column
    >>> With('ancestors', recursive=True).as_(
    ...     Union(
    ...         Select('*', table='folks', where=Eq('name', 'Alex')),
    ...         Select(
    ...             'f.*',
    ...             table=[
    ...                 Table("folks", alias="f"),
    ...                 Table("ancestors", alias="a")
    ...             ],
    ...             where=Eq("f.id", Column("a.parent_id"))
    ...         )
    ...     )
    ... ).select(Select(
    ...     'id', 'name', 'parent_id',
    ...     table='ancestors'
    ... ))
    ... "WITH RECURSIVE `ancestors` AS ("
    ...    "SELECT `id`, `name`, `parent_id` FROM `folks` WHERE `name` = %s "
    ...    "UNION SELECT `f`.`*` FROM `folks` AS `f`, `ancestors` AS `a` WHERE `f`.`id` = `a`.`parent_id`"
    ... ") SELECT `id`, `name`, `parent_id` FROM `ancestors`"
    ... ["Alex"]
    """

    def __init__(
        self,
        name: str | Table,
        columns: Collection[str | Column] | None = None,
        cte: Select | Union | None = None,
        select: Select | None = None,
        dialect: SQLDialect | None = None,
        *,
        recursive: bool = False,
    ) -> None:
        """
        :param name: Name of the resulting table produced by the CTE.
        :param columns: Optional list of column names the table should have.
        :param cte: Expression to create the table. This can be Select or Union.
        :param select: Select statement that will be executed after the CTE table has been constructed. Can use the CTE table.
        :param dialect: SQL dialect to use for the statement. If not provided, the default dialect will be used.
        :param recursive: If True, the CTE will be recursive. This means that the CTE can refer to itself in its own definition.
        """

        super().__init__(dialect=dialect)

        self._name = name if isinstance(name, Table) else Table(name)
        self._columns = [column if isinstance(column, Column) else Column(column) for column in columns] if columns else None
        self._recursive = recursive
        self._cte = cte
        self._select = select

    def as_(self, select: Select | Union) -> Self:
        """
        Specify the definition of the CTE table. This can be a Select or Union statement, that will produce the results for the
        CTE.
        :param select: Statement.
        """

        self._cte = select
        return self

    def select(self, select: Select) -> Self:
        """
        Select that will be executed after the CTE has been constructed.
        :param select: Select to execute.
        """

        self._select = select
        return self

    def __str__(self) -> str:
        """Return serialized CTE string."""

        if not self._cte:
            raise AttributeError("Missing CTE part of the select.")

        with self.dialect:
            columns = "" if not self._columns else f" ({', '.join(map(str, self._columns))})"

            return f"WITH {'RECURSIVE ' if self._recursive else ''}{self._name}{columns} AS ({self._cte}) {self._select}"

    def __bool__(self) -> bool:
        return self._cte is not None and bool(self._cte) and self._select is not None and bool(self._select)

    @property
    def args(self) -> list[Any]:
        """Arguments of the CTE statement."""

        args = [*self._name.args]

        if self._columns:
            for column in self._columns:
                args += column.args

        args.extend(
            [
                *(self._cte.args if self._cte else []),
                *(self._select.args if self._select else []),
            ]
        )

        return args



"""LIMIT statement"""

from __future__ import annotations

from typing import Any, Self, overload

from sqlfactory.statement import ConditionalStatement, Statement


class Limit(ConditionalStatement, Statement):
    """
    `LIMIT` statement

    Examples:

    ```python
    Limit()                   # => No LIMIT statement
    Limit(10)                 # => LIMIT 10
    Limit(5, 10)              # => LIMIT 5, 10
    Limit(offset=3)           # AttributeError, offset cannot be used without limit.
    Limit(offset=3, limit=5)  # => LIMIT 3, 5
    ```
    """

    @overload
    def __init__(self) -> None:
        """No LIMIT statement"""

    @overload
    def __init__(self, limit: int, /) -> None:
        """
        Just a LIMIT statement without offset
        :param limit: Number of returned rows
        """

    @overload
    def __init__(self, /, offset: int | None, limit: int | None) -> None:
        """
        LIMIT statement with both offset and limit
        :param offset: Pagination offset (how many rows to skip before returning result)
        :param limit: Number of returned rows
        """

    def __init__(  # type: ignore[misc]
        self, offset_or_limit: int | None = None, /, limit: int | None = None, *, offset: int | None = None
    ) -> None:
        """
        LIMIT statement
        :param offset_or_limit: Pagination offset, or limit if second argument is None. Only as positional argument.
        :param limit: Number of returned rows.
        :param offset: Optional keyword argument for specifying offset.
        """

        super().__init__()

        if offset_or_limit is not None and offset is not None and limit is not None:
            raise AttributeError("Unable to specify both positional argument offset and keyword argument offset.")

        if offset is not None and offset_or_limit is None and limit is None:
            raise AttributeError("Cannot use only offset without limit.")

        if limit is None:
            limit = offset_or_limit
            offset_or_limit = offset

        if offset is not None:
            offset_or_limit = offset

        self.offset = offset_or_limit
        """Specified pagination offset"""

        self.limit = limit
        """Specified number of returned rows"""

    def __str__(self) -> str:
        if self.offset is not None:
            return f"LIMIT {self.dialect.placeholder}, {self.dialect.placeholder}"

        if self.limit is not None:
            return f"LIMIT {self.dialect.placeholder}"

        return ""

    def __bool__(self) -> bool:
        """Return True if statement should be included in query, False otherwise."""
        return self.offset is not None or self.limit is not None

    @property
    def args(self) -> list[int]:
        """Argument values of the limit statement"""
        if self.offset is not None and self.limit is not None:
            return [self.offset, self.limit]

        if self.limit is not None:
            return [self.limit]

        return []


class WithLimit:
    """
    Mixin to provide LIMIT support for query generator.

    Usage:

    >>> Select().limit(10)  # SELECT ... LIMIT 10
    >>> Select().limit(5, 10)  # SELECT ... LIMIT 5, 10
    >>> Select().limit(limit=10, offset=5)  # SELECT ... LIMIT 5,10
    >>> Select().limit(Limit(...))
    >>> Select(..., limit=Limit(...))

    """

    def __init__(self, *args: Any, limit: Limit | None = None, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._limit = limit

    @overload
    def limit(self, limit: Limit | None, /) -> Self:
        """
        Limit statement
        :param limit: Instance of Limit
        """

    @overload
    def limit(self, limit: int | None, /) -> Self:
        """
        Limit statement
        :param limit: Number of returned rows
        """

    @overload
    def limit(self, offset: int | None, limit: int | None) -> Self:
        """
        Limit statement
        :param offset: Pagination offset (how many rows to skip before returning result)
        :param limit: Number of returned rows
        """

    def limit(  # type: ignore[misc]
        self, offset_or_limit: int | Limit | None = None, /, limit: int | None = None, *, offset: int | None = None
    ) -> Self:
        """
        Limit statement

        Usage:

        >>> Select(...).limit(10)  # SELECT ... LIMIT 10
        >>> Select(...).limit(5, 10)
        >>> Select(...).limit(limit=10, offset=5)
        >>> Select(...).limit(Limit(...))

        :param offset_or_limit: Positional-only argument as first argument to the LIMIT SQL statement. Can also be instance
            of `Limit`.

            If there is no `limit` argument, the first argument will be used as limit. If there is `limit` argument,
            the first argument will be used as offset. This weird behavior is to mimic SQL's LIMIT syntax.

        :param limit: Limit argument, can be passed as kwarg or as second positional argument.
        :param offset: Keyword-only argument to explicitly specify offset.
        """
        if self._limit is not None:
            raise AttributeError("Limit has already been specified.")

        if isinstance(offset_or_limit, Limit):
            self._limit = offset_or_limit

            if limit is not None or offset is not None:
                raise AttributeError("When passing Limit instance as first argument, second argument should not be passed.")

        else:
            if offset_or_limit is not None:
                if offset is not None:
                    raise AttributeError("Unable to specify both positional argument offset and keyword argument offset.")

                self._limit = Limit(offset_or_limit, limit)

            else:
                self._limit = Limit(offset=offset, limit=limit)

        return self

    @overload
    def LIMIT(self, limit: Limit | None, /) -> Self:
        # pylint: disable=invalid-name
        """Alias for `WithLimit.limit()` to be more SQL-like with all capitals."""

    @overload
    def LIMIT(self, limit: int | None, /) -> Self:
        # pylint: disable=invalid-name
        """Alias for `WithLimit.limit()` to be more SQL-like with all capitals."""

    @overload
    def LIMIT(self, offset: int | None, limit: int | None) -> Self:
        # pylint: disable=invalid-name
        """Alias for `WithLimit.limit()` to be more SQL-like with all capitals."""

    def LIMIT(  # type: ignore[misc]
        self, offset_or_limit: int | Limit | None, /, limit: int | None = None, *, offset: int | None = None
    ) -> Self:
        # pylint: disable=invalid-name
        """Alias for `WithLimit.limit()` to be more SQL-like with all capitals."""
        return self.limit(offset_or_limit, limit, offset=offset)  # type: ignore[call-overload]



"""Conditions for WHERE, ON, HAVING clauses in SQL statements."""

from sqlfactory.condition.base import And, Condition, ConditionBase, Or
from sqlfactory.condition.between import Between, NotBetween
from sqlfactory.condition.in_condition import In, NotIn
from sqlfactory.condition.like import Like, NotLike
from sqlfactory.condition.rlike import NotRLike, RLike
from sqlfactory.condition.simple import (
    Eq,
    Equals,
    Ge,
    GreaterThan,
    GreaterThanOrEquals,
    Gt,
    Le,
    LessThan,
    LessThanOrEquals,
    Lt,
    Ne,
    NotEquals,
)

__all__ = [
    "And",
    "Between",
    "Condition",
    "ConditionBase",
    "Eq",
    "Equals",
    "Ge",
    "GreaterThan",
    "GreaterThanOrEquals",
    "Gt",
    "In",
    "Le",
    "LessThan",
    "LessThanOrEquals",
    "Like",
    "Lt",
    "Ne",
    "NotBetween",
    "NotEquals",
    "NotIn",
    "NotLike",
    "NotRLike",
    "Or",
    "RLike",
]



"""Column / Statement aliasing support"""

from typing import Any

from sqlfactory.entities import Column, ColumnArg
from sqlfactory.statement import Statement


class Aliased(Statement):
    """
    Aliased generic statement. Only to be used in SELECT statement, where AS statement is only valid.

    Usage:

    >>> from sqlfactory import Aliased
    >>> from sqlfactory.func.agg import Count
    >>> Aliased(Count('*'), alias='count')
    >>> "COUNT(*) AS `count`"

    >>> from sqlfactory import Select, Aliased
    >>> from sqlfactory.func.agg import Count
    >>> Select(Aliased(Count('*'), alias='count'), table='orders')
    >>> 'SELECT COUNT(*) AS `count` FROM `orders`'
    """

    def __init__(self, statement: Statement | ColumnArg, alias: str | None = None) -> None:
        """
        :param statement: Statement to be aliased
        :param alias: Alias of the statement.
        """
        super().__init__()
        self._statement = statement if isinstance(statement, Statement) else Column(statement)

        self.alias = alias
        """Alias of the statement"""

    def __str__(self) -> str:
        if self.alias is None:
            return str(self._statement)

        from sqlfactory.select import Select  # pylint: disable=import-outside-toplevel, cyclic-import

        if isinstance(self._statement, Select):
            return f"({self._statement!s}) AS `{self.alias}`"

        return f"{self._statement!s} AS `{self.alias}`"

    @property
    def args(self) -> list[Any]:
        """Argument values of the aliased statement"""
        return self._statement.args

    def __getattr__(self, name: str) -> Any:
        """Proxy to access attributes of inner (non-aliased) statement."""
        return getattr(self._statement, name)


class SelectColumn(Aliased):
    """
    Aliased column. Shortcut for `Aliased(Column(column), alias)`

    Usage:

    >>> from sqlfactory import Select, SelectColumn
    >>> Select(SelectColumn("table.column", alias="otherColumn"), table="table")
    >>> "SELECT `table`.`column` AS `otherColumn` FROM `table`"

    >>> from sqlfactory import Select, SelectColumn
    >>> Select(SelectColumn("table.column", distinct=True), table="table")
    >>> "SELECT DISTINCT `table`.`column` FROM `table`"

    >>> from sqlfactory import Select, SelectColumn
    >>> Select(SelectColumn("table.column", alias="otherColumn", distinct=True), table="table")
    >>> "SELECT DISTINCT `table`.`column` AS `otherColumn` FROM `table`"
    """

    def __init__(self, column: ColumnArg, alias: str | None = None, distinct: bool = False):
        """
        :param column: Column to be selected
        :param alias: Optional alias of the column
        :param distinct: Whether to select only distinct values (prepend DISTINCT to column name)
        """
        super().__init__(column, alias)
        self.distinct = distinct
        """Whether to select only distinct values (adds DISTINCT to the column selector)."""

    def __str__(self) -> str:
        if self.distinct:
            return f"DISTINCT {super().__str__()}"

        return super().__str__()



"""Column list for usage in SELECT statement."""

from __future__ import annotations

from typing import Any, Iterable, Self

from sqlfactory.entities import Column, ColumnArg
from sqlfactory.statement import Statement


class ColumnList(Statement, list[Statement]):  # type: ignore[misc]
    """
    Unique(ish) set of columns to be used in SELECT statement. This class is usefull when you want to dynamically build
    list of selected columns and you don't want to have duplicate selects in the list.

    Usage:

    >>> c = ColumnList(["column1", "column2", "column3"])
    >>> c.append("column4")  # append() works, as with any other list.
    >>> c.add("column5")  # add() works too.
    >>> c.append("column1")  # this will not be added, because column1 is already in the list.
    >>>
    >>> print(c)
    >>> "[`column1`, `column2`, `column3`, `column4`, `column5`]"
    >>>
    >>> Select(select=c, table="table")
    >>> "SELECT `column1`, `column2`, `column3`, `column4`, `column5` FROM `table`"

    You can also use any other statement:

    >>> c = ColumnList()
    >>> c.append(If(Eq("column", True), "column2", "column3"))
    >>> print(c, c.args)
    >>> "[IF(`column` = %s, `column2`, `column3`)], [True]"

    Few more examples:

    >>> ColumnList([
    >>>    "c1"
    >>>    Column("c2"),
    >>>    Count("*"),
    >>>    SelectColumn("c3", alias="otherColumn"),
    >>>    Aliased(If(Eq("column", True), "column2", "column3"), alias="condition")
    >>>    Table("t").id,
    >>> ])
    >>> "[`c1`, `c2`, COUNT(*), `c3` AS `otherColumn`, IF(`column` = %s, `column2`, `column3`) AS `condition`, `t`.`id`]"
    """

    def __init__(self, iterable: Iterable[Statement | ColumnArg] | None = None) -> None:
        """
        :param iterable: Initial content of the column list. Can be either string, which is parsed into
        `[[database.]table.]column` tuple and then escaped properly. Or it can be a `Column` instance, which is added
        directly to the list. Any other statement is added as is.
        """
        if iterable:
            super().__init__([Column(i) if not isinstance(i, Statement) else i for i in iterable])
        else:
            super().__init__()

    def __contains__(self, other: Statement) -> bool:  # type: ignore[override]
        """
        Checks whether the column is already in the list. Handles correctly parsing of columns, so Column("c1") == "c1".
        """
        if not isinstance(other, Statement):
            raise AttributeError("ColumnList can only contain Statement objects.")

        for item in self:
            if str(item) == str(other) and item.args == other.args:
                return True

        return False

    def add(self, element: Statement | str) -> Self:
        """
        Add new columns to the set. If the column is already in the list, it is not added again.

        :param element: New select column to be added. Can be either string (which is treated as `Column` automatically),
            `Column` instance or any generic statement such as function call (`Count('*')` for example).
        """
        return self.append(element)

    def append(self, element: Statement | str) -> Self:  # type: ignore[override]
        """
        Add new columns to the set. If the column is already in the list, it is not added again.

        This is alias for `ColumnList.add()`.

        :param element: New select column to be added. Can be either string (which is treated as `Column` automatically),
            `Column` instance or any generic statement such as function call (`Count('*')` for example).
        """
        if not isinstance(element, Statement):
            element = Column(element)

        if element not in self:
            super().append(element)

        return self

    def update(self, iterable: Iterable[Statement | str]) -> Self:
        """
        Add multiple new columns to the set.
        :param iterable: New select columns to be added. Can be either string (which is treated as `Column` automatically),
            `Column` instance or any generic statement such as function call (`Count('*')` for example). Added columns
            are checked for uniqueness and duplicates are automatically skipped.
        """
        for item in iterable:
            self.add(item)

        return self

    def __str__(self) -> str:
        """
        Returns string representation of the column list usable for SELECT statement building.
        """
        return ", ".join(map(str, self))

    def __repr__(self) -> str:
        return "[" + ", ".join(map(repr, self)) + "]"

    @property
    def args(self) -> list[Any]:
        """Argument values of the column list statement."""
        out = []

        for item in self:
            if isinstance(item, Statement):
                out.extend(item.args)

        return out



"""
SQLFactory is a Python library for building SQL queries in a programmatic way.

Quick start:

## SELECT ...

Main class of interest: `Select`

### Quick example

```python
from sqlfactory import Select, Eq

query = Select("id", "name", table="products", where=Eq("enabled", True))
```

Adding another WHERE condition:

```python
query.where(Eq("id", 1))
```

**Note:** Multiple where method calls are chained using AND operator. If you need to use OR operator, you can use `Or` class.

**Note:** `Eq` and other conditions expects column name as first argument and value as second argument. If you need to compare
column with another column, you can use `Column` class as a second argument.

**Note:** Method chaining modifies the original statement in place, it does not return copy.

#### Ordering

```python
query.order_by("id", Direction.ASC)  # ORDER BY `id` ASC
```

#### Limit

```python
query.limit(10)     # LIMIT 10
query.limit(5, 10)  # LIMIT 5, 10
```

#### Group by

```python
query.group_by("id", "name", "create_date")  # GROUP BY `id`, `name`, `create_date`

query.having(Count("id") > 2)  # HAVING COUNT(`id`) > 2
```

**Note:** Multiple calls to `Select.group_by()` will combine the columns in the GROUP BY clause in the order the methods were
called.

**Note:** Multiple calls to `Select.having()` will combine the conditions using AND operator, same as `Select.where()`.

## INSERT ...

Main class of interest: `Insert`

Quick example:

```python
from sqlfactory import Insert

query = Insert.into("products")("id", "name").values(
    (1, "Product 1"),
    (2, "Product 2"),
)
```

**Note:** Multiple calls to `Insert.values()` will add more rows to the INSERT statement.

```python
from sqlfactory import Values

query.on_duplicate_key_update(name=Values("name"))   # ON DUPLICATE KEY UPDATE `name` = VALUES(`name`)
```

## UPDATE ...

Main class of interest: `Update`

Quick example:

```python
from sqlfactory import Update, Eq

query = Update("products", set={"enabled": False}, where=Eq("id", 1))

query.set("another_column", "value")  # SET `another_column` = %s
```

**Note:** Multiple calls to `Update.set()` will add more columns to the UPDATE statement.

## DELETE ...

Main class of interest: `Delete`

Quick example:

```python
from sqlfactory import Delete, Eq

query = Delete("products", where=Eq("enabled", False))
```

## Features

- `SELECT` statements
  - Variable select columns
  - Multiple tables
  - Subselects
  - Joins
  - Where condition
  - Group by
  - Having
  - Ordering
  - Limit
  - Common Table Expressions (CTE)
- `INSERT` statements
  - Multiple rows at once
  - Replace
  - On duplicate key update (MySQL / MariaDB specific)
  - `INSERT ... SELECT`
- `UPDATE` statements
  - Variable set of columns to change
  - Where condition
  - Limit
- `DELETE` statements
  - Join
  - Subselects
  - Where condition
  - Order
  - Limit
- `UNION` [`ALL`], `EXCEPT` [`ALL`], `INTERSECT` [`ALL`]
- Multiple SQL dialects support - MySQL / MariaDB, SQLite, PostgreSQL, custom.
  - Can support numeric placeholders (`$1`, `$2`, ...) in custom implementation.

## Basic concepts

- Each component of SQL statement is represented by its own class.
- Each (part of) SQL statement can be used individually or combined with other parts to create more complex SQL statements.
- Each part of statement bundles string representation of given portion of SQL including placeholders with placeholder values.
  These parts are then combined to construct more complex SQL statements up to the executable ones (SELECT, INSERT, UPDATE,
  DELETE, ...).
- When there is no expected return value, `self` is always returned to allow chaining of methods. Chaining of method does not
  have given order, methods can be chained as needed. For example, WHERE does not have to be before LIMIT for SELECT. It is up
  to the library to figure out correct order of parts when building final query.
- `statement.__str__()` method (`str(statement)`) is used to get the final SQL statement.
- `statement.args` property is used to get the values of placeholders.
- `statement.execute()` function is used to execute the statement that could be executed on top of DB-API 2.0 compatible cursor
  (or any object that has `execute()` method taking string as SQL and variable arguments or tuple of arguments as placeholder
  values).
- It does not matter whether you are in async or sync code, or what database you are using. SQLFactory is database-agnostic and
    does not care about the underlying database. It only cares about building the SQL statement.
- SQLFactory is not an ORM. It does not map objects to tables or rows. By design.
- SQLFactory utilizes modern Python syntax and type hints, requiring Python 3.11+. It is fully type-safe and works well with
  IDE code completion.
- Developer who knows SQL should understand query written using SQLFactory without explicitly knowing SQLFactory API. Library
  tries to mimic SQL syntax as much as possible.

## SQL Dialect

SQLFactory supports multiple SQL dialects. By default, it uses MySQL / MariaDB dialect, but it can be changed to SQLite,
PostgreSQL, Oracle or custom dialect. Custom dialect can be used to support any database that supports standard SQL, but uses
different placeholders than `%s` (like `$1`, `$2`, ...).

```python
from sqlfactory import PostgreSQLDialect, Statement, Select, Eq, In

# Change the default dialect to PostgreSQL
Statement.default_dialect = PostgreSQLDialect()

# Now, all statements will use PostgreSQL dialect
select = Select("column1", "column2", table="books").where(Eq("column1", "value") & In("column2", [1, 2, 3]))
str(select)  # SELECT "column1", "column2" FROM "books" WHERE "column1" = %s AND "column2" IN (%s, %s, %s)
```

Dialects where placeholders must be numbered are also supported as can be seen in the example `OracleSQLDialect`:

```python
from sqlfactory import OracleSQLDialect, Statement, Select, Eq, In

# Change the default dialect to PostgreSQL
Statement.default_dialect = OracleSQLDialect()

# Now, all statements will use PostgreSQL dialect
select = Select("column1", "column2", table="books").where(Eq("column1", "value") & In("column2", [1, 2, 3]))
str(select)  # SELECT "column1", "column2" FROM "books" WHERE "column1" = :1 AND "column2" IN (:2, :3, :4)
```

However, the library is mainly tested with MySQL / MariaDB dialect, so there might be some issues with other dialects, such
as slightly different syntax or unsupported features. If you find any issue, please report it.

"""

from sqlfactory import execute, func, mixins  # exported submodules
from sqlfactory.condition import (
    And,
    Between,
    Condition,
    ConditionBase,
    Eq,
    Equals,
    Ge,
    GreaterThan,
    GreaterThanOrEquals,
    Gt,
    In,
    Le,
    LessThan,
    LessThanOrEquals,
    Like,
    Lt,
    Ne,
    NotBetween,
    NotEquals,
    NotIn,
    NotLike,
    NotRLike,
    Or,
    RLike,
)
from sqlfactory.delete import DELETE, Delete
from sqlfactory.dialect import MySQLDialect, OracleSQLDialect, PostgreSQLDialect, SQLDialect, SQLiteDialect
from sqlfactory.entities import Column, Table
from sqlfactory.insert import INSERT, Insert, Values
from sqlfactory.mixins import Direction, Limit, Order
from sqlfactory.select import (
    EXCEPT,
    EXCEPT_ALL,
    EXCEPT_DISTINCT,
    INTERSECT,
    INTERSECT_ALL,
    INTERSECT_DISTINCT,
    SELECT,
    UNION,
    UNION_ALL,
    UNION_DISTINCT,
    Aliased,
    ColumnList,
    CrossJoin,
    Except,
    ExceptAll,
    ExceptDistinct,
    InnerJoin,
    Intersect,
    IntersectAll,
    IntersectDistinct,
    Join,
    LeftJoin,
    LeftOuterJoin,
    RightJoin,
    RightOuterJoin,
    Select,
    SelectColumn,
    Union,
    UnionAll,
    UnionDistinct,
    With,
)
from sqlfactory.statement import Raw, Statement, Value
from sqlfactory.update import UPDATE, Update, UpdateColumn

__all__ = [  # noqa: RUF022
    "DELETE",
    "INSERT",
    "SELECT",
    "UPDATE",
    "Aliased",
    "And",
    "Between",
    "Column",
    "ColumnList",
    "Condition",
    "ConditionBase",
    "CrossJoin",
    "Delete",
    "Direction",
    "Eq",
    "Equals",
    "Except",
    "ExceptAll",
    "ExceptDistinct",
    "EXCEPT",
    "EXCEPT_ALL",
    "EXCEPT_DISTINCT",
    "Ge",
    "GreaterThan",
    "GreaterThanOrEquals",
    "Gt",
    "In",
    "InnerJoin",
    "Insert",
    "Intersect",
    "IntersectAll",
    "IntersectDistinct",
    "INTERSECT",
    "INTERSECT_ALL",
    "INTERSECT_DISTINCT",
    "Join",
    "Le",
    "LeftJoin",
    "LeftOuterJoin",
    "LessThan",
    "LessThanOrEquals",
    "Like",
    "Limit",
    "Lt",
    "Ne",
    "NotEquals",
    "NotIn",
    "NotLike",
    "NotBetween",
    "NotRLike",
    "Or",
    "Order",
    "Raw",
    "RightJoin",
    "RightOuterJoin",
    "RLike",
    "Select",
    "SelectColumn",
    "Statement",
    "Table",
    "Union",
    "UnionAll",
    "UnionDistinct",
    "UNION",
    "UNION_ALL",
    "UNION_DISTINCT",
    "Update",
    "UpdateColumn",
    "Value",
    "Values",
    "With",
    "SQLDialect",
    "MySQLDialect",
    "SQLiteDialect",
    "PostgreSQLDialect",
    "OracleSQLDialect",
    "execute",
    "func",
    "mixins",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from contextvars import ContextVar, Token
from types import TracebackType
from typing import ClassVar, Optional, Self, Type


class SQLDialect(ABC):
    """
    Base class for representing SQL dialect for building SQL queries.

    This class is used to configure SQL dialect for building SQL queries. It is used to configure placeholders for
    arguments and quote characters for table and column names.

    Usage:

    >>> from sqlfactory import PostgreSQLDialect, SQLiteDialect, Statement, Select, Eq
    >>>
    >>> # Set default dialect for whole library.
    >>> Statement.default_dialect = PostgreSQLDialect()
    >>>
    >>> sel = Select("a", table="table", where=Eq("id", 1))
    >>> str(sel)
    SELECT "a" FROM "table" WHERE "id" = %s
    >>>
    >>> # Overwrite dialect for this particular statement
    >>> sel = Select("a", table="table", where=Eq("id", 1), dialect=SQLiteDialect())
    >>> str(sel)
    SELECT `a` FROM `table` WHERE `id` = ?
    >>>
    >>> # Use dialect for all statements in the context.
    >>> with SQLiteDialect():
    ...   sel = Select("a", table="table", where=Eq("id", 1))
    ...   str(sel)  # SELECT `a` FROM `table` WHERE `id` = ?
    ...
    ...   # Explicit dialect always has precedence over context.
    ...   sel = Select(dialect=PostgreSQLDialect())
    ...   str(sel)  # SELECT "a" FROM "table" WHERE "id" = %s


    Implementing custom dialect:

    >>> from sqlfactory import SQLDialect, In
    >>>
    >>> class CustomDialect(SQLDialect):
    ...     @property
    ...     def placeholder(self) -> str:
    ...         return "$hello"
    ...
    ...     @property
    ...     def quote(self) -> str:
    ...         return '*'
    >>>
    >>> with CustomDialect():
    ...    str(In("column", [1, 2, 3]))
    ...
    '*column* IN ($hello, $hello, $hello)'

    """

    dialect_context: ClassVar[ContextVar[Optional[SQLDialect]]] = ContextVar("SQLDialect", default=None)
    """
    Context variable holding currently set SQL dialect.
    """

    def __init__(self) -> None:
        self._token: list[Token[SQLDialect | None]] = []

    @property
    @abstractmethod
    def placeholder(self) -> str:
        """Placeholder for arguments in SQL query."""

    @property
    @abstractmethod
    def quote(self) -> str:
        """Quote character for table and column names."""

    def __enter__(self) -> Self:
        self._token.append(self.dialect_context.set(self))
        return self

    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_stack: Optional[TracebackType]
    ) -> None:
        self.dialect_context.reset(self._token.pop())


class MySQLDialect(SQLDialect):
    """
    MySQL dialect for SQLFactory.
    """

    @property
    def placeholder(self) -> str:
        return "%s"

    @property
    def quote(self) -> str:
        return "`"


class SQLiteDialect(SQLDialect):
    """
    SQLite dialect for SQLFactory.
    """

    @property
    def placeholder(self) -> str:
        return "?"

    @property
    def quote(self) -> str:
        return "`"


class PostgreSQLDialect(SQLDialect):
    """
    PostgreSQL dialect for SQLFactory.
    """

    @property
    def placeholder(self) -> str:
        return "%s"

    @property
    def quote(self) -> str:
        return '"'


class OracleSQLDialect(SQLDialect):
    """
    Oracle SQL dialect for SQLFactory. With indexed placeholders.
    """

    def __init__(self, initial_index: int = 1) -> None:
        """
        :param initial_index: Index where to start counting placeholders.
        """
        super().__init__()
        self._index = initial_index
        self._initial_index = initial_index

    @property
    def placeholder(self) -> str:
        index = self._index
        self._index += 1
        return f":{index}"

    @property
    def quote(self) -> str:
        return '"'

    def __enter__(self) -> Self:
        self._index = self._initial_index
        return super().__enter__()
